@@ -1,13 +1,13 @@
 #include <algorithm>
+#include <chrono>
 #include <cstdint>
 #include <cstdlib>
 #include <gtest/gtest.h>
 
-#include <chrono>
 #include <stdexcept>
 #include <string>
 
-#include <cpr/cpr.h>
+#include "cpr/cpr.h"
 #include <curl/curl.h>
 #include <vector>
 
@@ -21,10 +21,46 @@ static HttpServer* server = new HttpServer();
 std::chrono::milliseconds sleep_time{50};
 std::chrono::seconds zero{0};
 
-bool write_data(std::string /*data*/, intptr_t /*userdata*/) {
+bool write_data(const std::string_view& /*data*/, intptr_t /*userdata*/) {
     return true;
 }
 
+TEST(SessionGetTests, GetMultipleTimes) {
+    Url url{server->GetBaseUrl() + "/hello.html"};
+    Session session;
+    session.SetUrl(url);
+    std::string expected_text{"Hello world!"};
+
+    for (size_t i = 0; i < 100; i++) {
+        Response response = session.Get();
+        EXPECT_EQ(expected_text, response.text);
+        EXPECT_EQ(url, response.url);
+        EXPECT_EQ(std::string{"text/html"}, response.header["content-type"]);
+        EXPECT_EQ(200, response.status_code);
+        EXPECT_EQ(ErrorCode::OK, response.error.code);
+    }
+}
+
+TEST(SessionPostTests, PostMultipleTimes) {
+    Url url{server->GetBaseUrl() + "/url_post.html"};
+    Session session;
+    session.SetUrl(url);
+    session.SetPayload({{"x", "5"}});
+    std::string expected_text{
+            "{\n"
+            "  \"x\": 5\n"
+            "}"};
+
+    for (size_t i = 0; i < 100; i++) {
+        Response response = session.Post();
+        EXPECT_EQ(expected_text, response.text);
+        EXPECT_EQ(url, response.url);
+        EXPECT_EQ(std::string{"application/json"}, response.header["content-type"]);
+        EXPECT_EQ(201, response.status_code);
+        EXPECT_EQ(ErrorCode::OK, response.error.code);
+    }
+}
+
 TEST(RedirectTests, TemporaryDefaultRedirectTest) {
     Url url{server->GetBaseUrl() + "/temporary_redirect.html"};
     Session session;
@@ -226,7 +262,7 @@ TEST(MultipleGetTests, HeaderChangeMultipleGetTest) {
         EXPECT_EQ(200, response.status_code);
         EXPECT_EQ(ErrorCode::OK, response.error.code);
     }
-    session.SetHeader(Header{{"key", "value"}});
+    session.SetHeader(Header{{"key", "value"}, {"lorem", "ipsum"}});
     {
         Response response = session.Get();
         std::string expected_text{"Header reflect GET"};
@@ -234,6 +270,19 @@ TEST(MultipleGetTests, HeaderChangeMultipleGetTest) {
         EXPECT_EQ(url, response.url);
         EXPECT_EQ(std::string{"text/html"}, response.header["content-type"]);
         EXPECT_EQ(std::string{"value"}, response.header["key"]);
+        EXPECT_EQ(std::string{"ipsum"}, response.header["lorem"]);
+        EXPECT_EQ(200, response.status_code);
+        EXPECT_EQ(ErrorCode::OK, response.error.code);
+    }
+    Header& headerMap = session.GetHeader();
+    headerMap.erase("key");
+    {
+        Response response = session.Get();
+        std::string expected_text{"Header reflect GET"};
+        EXPECT_EQ(expected_text, response.text);
+        EXPECT_EQ(url, response.url);
+        EXPECT_EQ(std::string{"text/html"}, response.header["content-type"]);
+        EXPECT_EQ(std::string{"ipsum"}, response.header["lorem"]);
         EXPECT_EQ(200, response.status_code);
         EXPECT_EQ(ErrorCode::OK, response.error.code);
     }
@@ -622,6 +671,24 @@ TEST(MultipartTests, SetMultipartValueTest) {
     EXPECT_EQ(ErrorCode::OK, response.error.code);
 }
 
+TEST(MultipartTests, SetMultipartVectorPartsTest) {
+    Url url{server->GetBaseUrl() + "/form_post.html"};
+    Session session;
+    session.SetUrl(url);
+    Multipart multipart{std::vector<Part>{Part{"x", "5"}}};
+    session.SetMultipart(multipart);
+    Response response = session.Post();
+    std::string expected_text{
+            "{\n"
+            "  \"x\": \"5\"\n"
+            "}"};
+    EXPECT_EQ(expected_text, response.text);
+    EXPECT_EQ(url, response.url);
+    EXPECT_EQ(std::string{"application/json"}, response.header["content-type"]);
+    EXPECT_EQ(201, response.status_code);
+    EXPECT_EQ(ErrorCode::OK, response.error.code);
+}
+
 TEST(BodyTests, SetBodyTest) {
     Url url{server->GetBaseUrl() + "/url_post.html"};
     Session session;
@@ -711,8 +778,8 @@ TEST(CookiesTests, BasicCookiesTest) {
     Cookies res_cookies{response.cookies};
     std::string expected_text{"Basic Cookies"};
     cpr::Cookies expectedCookies{
-            {"SID", "31d4d96e407aad42", "127.0.0.1", false, "/", true, std::chrono::system_clock::from_time_t(3905119080)},
-            {"lang", "en-US", "127.0.0.1", false, "/", true, std::chrono::system_clock::from_time_t(3905119080)},
+            {"SID", "31d4d96e407aad42", "127.0.0.1", false, "/", true, HttpServer::GetCookieExpiresIn100HoursTimePoint()},
+            {"lang", "en-US", "127.0.0.1", false, "/", true, HttpServer::GetCookieExpiresIn100HoursTimePoint()},
     };
     EXPECT_EQ(std::string{"text/html"}, response.header["content-type"]);
     EXPECT_EQ(200, response.status_code);
@@ -938,9 +1005,34 @@ TEST(DifferentMethodTests, MultipleDeleteHeadPutGetPostTest) {
     Url url{server->GetBaseUrl() + "/header_reflect.html"};
     Url urlPost{server->GetBaseUrl() + "/post_reflect.html"};
     Url urlPut{server->GetBaseUrl() + "/put.html"};
+    Url urlMultipartPost{server->GetBaseUrl() + "/post_file_upload.html"};
     Session session;
     for (size_t i = 0; i < 10; ++i) {
         {
+            session.RemoveContent();
+            session.SetUrl(url);
+            Response response = session.Get();
+            std::string expected_text{"Header reflect GET"};
+            EXPECT_EQ(expected_text, response.text);
+            EXPECT_EQ(url, response.url);
+            EXPECT_EQ(200, response.status_code);
+            EXPECT_EQ(ErrorCode::OK, response.error.code);
+        }
+        {
+            session.RemoveContent();
+            session.SetUrl(urlMultipartPost);
+            std::string fileContentsBinary{"this is a binary payload"};
+            std::string fileExtension = ".myfile";
+            session.SetMultipart(cpr::Multipart{{"files", cpr::Buffer{fileContentsBinary.begin(), fileContentsBinary.end(), "myfile.jpg"}}, {"file_types", "[\"" + fileExtension + "\"]"}});
+            Response response = session.Post();
+            std::string expected_text{"{\n  \"files\": \"myfile.jpg=this is a binary payload\",\n  \"file_types\": \"[\".myfile\"]\"\n}"};
+            EXPECT_EQ(expected_text, response.text);
+            EXPECT_EQ(urlMultipartPost, response.url);
+            EXPECT_EQ(201, response.status_code);
+            EXPECT_EQ(ErrorCode::OK, response.error.code);
+        }
+        {
+            session.RemoveContent();
             session.SetUrl(url);
             Response response = session.Delete();
             std::string expected_text{"Header reflect DELETE"};
@@ -960,6 +1052,7 @@ TEST(DifferentMethodTests, MultipleDeleteHeadPutGetPostTest) {
             EXPECT_EQ(ErrorCode::OK, response.error.code);
         }
         {
+            session.RemoveContent();
             session.SetUrl(url);
             Response response = session.Get();
             std::string expected_text{"Header reflect GET"};
@@ -983,6 +1076,7 @@ TEST(DifferentMethodTests, MultipleDeleteHeadPutGetPostTest) {
             EXPECT_EQ(ErrorCode::OK, response.error.code);
         }
         {
+            session.RemoveContent();
             session.SetUrl(url);
             Response response = session.Head();
             std::string expected_text{"Header reflect HEAD"};
@@ -1019,20 +1113,33 @@ TEST(CurlHolderManipulateTests, CustomOptionTest) {
 TEST(LocalPortTests, SetLocalPortTest) {
     Url url{server->GetBaseUrl() + "/local_port.html"};
     Session session;
-    session.SetUrl(url);
-    std::uint16_t const local_port = 60252; // beware of HttpServer::GetPort when changing
-    std::uint16_t const local_port_range = 5000;
-    session.SetLocalPort(local_port);
-    session.SetLocalPortRange(local_port_range);
-    // expected response: body contains port number in specified range
-    // NOTE: even when trying up to 5000 ports there is the chance that all of them are occupied.
-    // It would be possible to also check here for ErrorCode::INTERNAL_ERROR but that somehow seems
-    // wrong as then this test would pass in case SetLocalPort does not work at all
-    // or in other words: we have to assume that at least one port in the specified range is free.
-    Response response = session.Get();
+    uint16_t local_port{0};
+    uint16_t local_port_range{0};
+    Response response;
+
+    // Try up to 10 times to get a free local port
+    for (size_t i = 0; i < 10; i++) {
+        session.SetUrl(url);
+        local_port = 40252 + (i * 100); // beware of HttpServer::GetPort when changing
+        local_port_range = 7000;
+        session.SetLocalPort(local_port);
+        session.SetLocalPortRange(local_port_range);
+        // expected response: body contains port number in specified range
+        // NOTE: even when trying up to 7000 ports there is the chance that all of them are occupied.
+        // It would be possible to also check here for ErrorCode::UNKNOWN_ERROR but that somehow seems
+        // wrong as then this test would pass in case SetLocalPort does not work at all
+        // or in other words: we have to assume that at least one port in the specified range is free.
+        response = session.Get();
+
+        if (response.error.code != ErrorCode::UNKNOWN_ERROR) {
+            break;
+        }
+    }
+
     EXPECT_EQ(200, response.status_code);
     EXPECT_EQ(ErrorCode::OK, response.error.code);
-    std::uint16_t port_from_response = std::strtoul(response.text.c_str(), nullptr, 10);
+    // NOLINTNEXTLINE(google-runtime-int)
+    unsigned long port_from_response = std::strtoul(response.text.c_str(), nullptr, 10);
     EXPECT_EQ(errno, 0);
     EXPECT_GE(port_from_response, local_port);
     EXPECT_LE(port_from_response, local_port + local_port_range);
@@ -1041,19 +1148,32 @@ TEST(LocalPortTests, SetLocalPortTest) {
 TEST(LocalPortTests, SetOptionTest) {
     Url url{server->GetBaseUrl() + "/local_port.html"};
     Session session;
-    session.SetUrl(url);
-    std::uint16_t const local_port = 60551; // beware of HttpServer::GetPort when changing
-    std::uint16_t const local_port_range = 5000;
-    session.SetOption(LocalPort(local_port));
-    session.SetOption(LocalPortRange(local_port_range));
-    // expected response: body contains port number in specified range
-    // NOTE: even when trying up to 5000 ports there is the chance that all of them are occupied.
-    // It would be possible to also check here for ErrorCode::INTERNAL_ERROR but that somehow seems
-    // wrong as then this test would pass in case SetOption(LocalPort) does not work at all
-    // or in other words: we have to assume that at least one port in the specified range is free.
-    Response response = session.Get();
+    uint16_t local_port{0};
+    uint16_t local_port_range{0};
+    Response response;
+
+    // Try up to 10 times to get a free local port
+    for (size_t i = 0; i < 10; i++) {
+        session.SetUrl(url);
+        local_port = 30252 + (i * 100); // beware of HttpServer::GetPort when changing
+        local_port_range = 7000;
+        session.SetOption(LocalPort(local_port));
+        session.SetOption(LocalPortRange(local_port_range));
+        // expected response: body contains port number in specified range
+        // NOTE: even when trying up to 7000 ports there is the chance that all of them are occupied.
+        // It would be possible to also check here for ErrorCode::UNKNOWN_ERROR but that somehow seems
+        // wrong as then this test would pass in case SetLocalPort does not work at all
+        // or in other words: we have to assume that at least one port in the specified range is free.
+        response = session.Get();
+
+        if (response.error.code != ErrorCode::UNKNOWN_ERROR) {
+            break;
+        }
+    }
+
     EXPECT_EQ(200, response.status_code);
     EXPECT_EQ(ErrorCode::OK, response.error.code);
+    // NOLINTNEXTLINE(google-runtime-int)
     unsigned long port_from_response = std::strtoul(response.text.c_str(), nullptr, 10);
     EXPECT_EQ(errno, 0);
     EXPECT_GE(port_from_response, local_port);
@@ -1073,7 +1193,7 @@ TEST(LocalPortTests, SetLocalPortTestOccupied) {
     session.SetLocalPort(server->GetPort());
     // expected response: request cannot be made as port is already occupied
     Response response = session.Get();
-    EXPECT_EQ(ErrorCode::INTERNAL_ERROR, response.error.code);
+    EXPECT_EQ(ErrorCode::INTERFACE_FAILED, response.error.code);
 }
 
 TEST(LocalPortTests, SetOptionTestOccupied) {
@@ -1083,7 +1203,7 @@ TEST(LocalPortTests, SetOptionTestOccupied) {
     session.SetOption(LocalPort(server->GetPort()));
     // expected response: request cannot be made as port is already occupied
     Response response = session.Get();
-    EXPECT_EQ(ErrorCode::INTERNAL_ERROR, response.error.code);
+    EXPECT_EQ(ErrorCode::INTERFACE_FAILED, response.error.code);
 }
 #endif // _WIN32
 
@@ -1281,7 +1401,7 @@ TEST(AsyncRequestsTests, AsyncGetMultipleTemporarySessionTest) {
 TEST(AsyncRequestsTests, AsyncGetMultipleReflectTest) {
     Url url{server->GetBaseUrl() + "/hello.html"};
     std::vector<AsyncResponse> responses;
-    for (size_t i = 0; i < 100; ++i) {
+    for (size_t i = 0; i < 10; ++i) {
         std::shared_ptr<Session> session = std::make_shared<Session>();
         session->SetUrl(url);
         session->SetParameters({{"key", std::to_string(i)}});
@@ -1503,6 +1623,18 @@ TEST(CallbackTests, PatchCallbackTest) {
     EXPECT_EQ(ErrorCode::OK, response.error.code);
 }
 
+TEST(CallbackTests, Move) {
+    auto session = Session();
+    session.SetDebugCallback(DebugCallback([](auto, auto, auto) {}));
+
+    auto use = +[](Session& s) {
+        s.SetUrl(server->GetBaseUrl());
+        s.Get();
+    };
+    use(session);
+}
+
+
 int main(int argc, char** argv) {
     ::testing::InitGoogleTest(&argc, argv);
     ::testing::AddGlobalTestEnvironment(server);
